-- Force exerted on the player, and minimum baseline speed of the shot
local MIN_FORCE_PUSH = 16*FRACUNIT
-- Shot travel speed is multiplied by this AFTER applying the minimum
local PUSH_SPEED_FACTOR = FRACUNIT/2
-- How many tics the shot lasts for when travelling through air
local FORCE_SHOT_LIFESPAN = 25
-- Base radius of rhe orb around the shot
local FORCE_BLAST_RADIUS = 48*FRACUNIT
local FORCE_BLAST_HEIGHT = FORCE_BLAST_RADIUS*2

freeslot(
"S_FORCESHOTSPAWN",
"S_FORCESHOT",
"S_FORCESHOTDIE",
"S_FORCESHOTDEAD",
"MT_FORCESHOT",
"S_FORCEBLAST",
"S_FORCEBLASTDIE",
"MT_FORCEBLAST"
)

-- Force shield activation
function A_KRForceShield(p)
	-- Spawn the force shot
	local forceshot = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_FORCESHOT)
	-- Orb remembers which player spawned it
	forceshot.shooter = p
	forceshot.orb.shooter = p
	-- Position the orb's center at half the player's height
	forceshot.z = p.mo.z + (p.mo.height/2)
	-- Tell the orb where its center is
	forceshot.orb.center = forceshot.z
	
	-- Direction the player is moving horizontally
	local momangle = R_PointToAngle2(0, 0, p.mo.momx, p.mo.momy)
	-- Speed the player is moving horizontally
	local speed = FixedHypot(abs(p.mo.momx), abs(p.mo.momy))
	
	-- Decide which direction the player wants to launch away from
	-- In simple mode or freeaim standard while holding the stick:
	if FixedHypot(abs(p.cmd.forwardmove), abs(p.cmd.sidemove)) > 5
	and ((p.pflags & PF_ANALOGMODE) or ((p.pflags & PF_DIRECTIONCHAR) and krework.freeaim(p)))
		-- Aim in the direction you're holding on the stick
		p.drawangle = p.krework.realInputAngle()
	else
		-- Otherwise, just go the way the character is already pointing
		p.drawangle = p.mo.angle
	end
	forceshot.angle = p.drawangle
	
	-- Scale launch speed down based on how far away from your direction of motion you're angled
	forceshot.speed = FixedMul(PUSH_SPEED_FACTOR,
								MIN_FORCE_PUSH + FixedMul(cos(momangle - forceshot.angle), speed))
	
	-- Fire the player and the force bubble in opposite directions
	P_InstaThrust(forceshot, forceshot.angle, forceshot.speed)
	P_InstaThrust(p.mo, forceshot.angle+ANGLE_180, MIN_FORCE_PUSH)
	p.mo.momz = 0
	
	S_StartSound(p.mo, sfx_ngskid)
end

-- Core hitbox for the force push orb
-- Force push core spawning
function A_KRForceShotSpawn(forceshot, var1, var2)
	-- The core is invisible
	forceshot.flags2 = $ | MF2_DONTDRAW
	
	-- Spawn the orb
	forceshot.orb = P_SpawnMobjFromMobj(forceshot, 0, 0, 0, MT_FORCEBLAST)
	-- The orb remembers what its core is
	forceshot.orb.core = forceshot
end
states[S_FORCESHOTSPAWN] = {action = A_KRForceShotSpawn,
	sprite = SPR_FORC,
	frame = A,
	tics = 1,
	var1 = 0,
	var2 = 0,
	nextstate = S_FORCESHOT
}

-- Force push core travelling
states[S_FORCESHOT] = {
	sprite = SPR_FORC,
	frame = A,
	tics = FORCE_SHOT_LIFESPAN,
	var1 = 0,
	var2 = 0,
	nextstate = S_FORCESHOTDIE
}

--Force push core dying
function A_KRForceShotDie(forceshot, var1, var2)
	forceshot.orb.state = S_RAISESTATE
end
states[S_FORCESHOTDIE] = {action = A_KRForceShotDie,
	sprite = SPR_FORC,
	frame = A,
	tics = 20,
	var1 = 0,
	var2 = 0,
	nextstate = S_DEATHSTATE
}

mobjinfo[MT_FORCESHOT] = {
	doomednum = -1,
	spawnstate = S_FORCESHOTSPAWN,
	spawnhealth = 1000,
	reactiontime = 0,
	radius = 1,
	height = 1,
	mass = 0,
	flags = MF_NOGRAVITY|MF_RUNSPAWNFUNC
}

-- Core's regular thinker
addHook("MobjThinker", function(forceshot)
	if forceshot.state == S_FORCESHOT
		-- Speed decays to 0 over the projectile's lifespan
		P_Thrust(forceshot, forceshot.angle + ANGLE_180, forceshot.speed / FORCE_SHOT_LIFESPAN)
	end
end, MT_FORCESHOT)

-- Core dies if it touches terrain
addHook("MobjMoveBlocked", function(FORCESHOT)
	if FORCESHOT.state == S_FORCESHOT
		FORCESHOT.state = S_FORCESHOTDIE
	end
end, MT_FORCESHOT)

-- Bullet blocking/visual/damage for the force push orb
-- Force orb's normal lifespan
function A_KRForceBlastTravel(forceblast, var1, var2)
	-- Orb tracks its core
	if forceblast.core
		P_TeleportMove(forceblast, forceblast.core.x, forceblast.core.y, forceblast.center-(forceblast.radius/2))
	end
end
states[S_FORCEBLAST] = {action = A_KRForceBlastTravel,
	sprite = SPR_FORC,
	frame = K|TR_TRANS50|FF_ANIMATE|FF_GLOBALANIM,
	tics = 1,
	var1 = 9,
	var2 = 1,
	nextstate = S_FORCEBLAST
}

-- Force orb dying
function A_KRForceBlastDeath(forceblast, var1, var2)
	-- Make the orb expand as it expires
	forceblast.destscale = 3*FRACUNIT
	forceblast.scalespeed = abs((forceblast.destscale - forceblast.scale) / forceblast.tics)
end
states [S_FORCEBLASTDIE] = {action = A_KRForceBlastDeath,
	sprite = SPR_FORC,
	frame = K|TR_TRANS50|FF_ANIMATE|FF_GLOBALANIM,
	tics = 20,
	var1 = 9,
	var2 = 1,
	nextstate = S_DEATHSTATE
}

mobjinfo[MT_FORCEBLAST] = {
	doomednum = -1,
	spawnstate = S_FORCEBLAST,
	spawnhealth = 1000,
	raisestate = S_FORCEBLASTDIE,
	reactiontime = 0,
	radius = FORCE_BLAST_RADIUS,
	height = FORCE_BLAST_HEIGHT,
	mass = 0,
	flags = MF_NOCLIPHEIGHT|MF_SHOOTABLE|MF_NOGRAVITY|MF_RUNSPAWNFUNC
}

-- Hook to tell the force orb to ignore bullets
addHook("ShouldDamage", function(forceblast, inflictor, source, damage, damagetype)
	-- Block bullets
	if inflictor != source
		return false
	end
end, MT_FORCEBLAST)

-- Orb collision/damage check
addHook("MobjMoveCollide", function(forceblast, touched)
	local p = forceblast.shooter
	
	if not p
	or not krework.checkCollisionHeight(forceblast, touched)
	or not (touched.flags & (MF_MONITOR|MF_ENEMY|MF_BOSS))
	or not (touched.flags & MF_SHOOTABLE)
	or (touched.flags & MF_MISSILE)
		return false
	end
	
	P_DamageMobj(touched, forceblast, p.mo, 1)
	
	return true
end, MT_FORCEBLAST)
