rawset(_G.krework, "tails", {})

-- SETTINGS
rawset(_G.krework.tails, "settings", {
-- Number of times you can boost per jump
BOOST_COUNT = 3,
-- Boost ascent speed
BOOST_UPSPEED = (53/20)*FRACUNIT,
-- Number of tics you spend boosting
FLY_BOOST_TIME = 7,
-- Number of tics you spend boosting when swimming
SWIM_BOOST_TIME = 4,
-- Minimum time between flight boosts
FLY_SPAMLOCK = 10,
-- How fast you fall when first hovering
FLY_HOVERSPEED = -(7/2) * FRACUNIT,
-- How fast you fall towards the end of your hover
FLY_TIRED_HOVERSPEED = -(11/2) * FRACUNIT,
-- How many tics it takes to hit the end of hover state
FLY_TIRED_TIME = 55,
-- How many tics it takes to drop out of the air entirely
FLY_END_TIME = 85,

-- Whether to resize the dust clouds produced by flight boosts
BOOST_SCALE_DUST = true,
-- Whether to color the dust clouds produced by flight boosts
BOOST_COLOR_DUST = false,
-- Colors for each stage of dust
BOOST_DUST_COLORS = {[2] = SKINCOLOR_YELLOW, [1] = SKINCOLOR_ORANGE, [0] = SKINCOLOR_RED},
BOOST_DUST_DEFAULT = SKINCOLOR_EMERALD
})

local set = krework.tails.settings

rawset(_G.krework.tails, "reset", function(mo)
    -- Reset how long Tails has been flying
    mo.krework.tails.flytime = 0
    -- Reset how many more times Tails can boost
    mo.krework.tails.flycount = set.BOOST_COUNT
end)

function A_KRTailsResetFlyCount(p)
	p.mo.krework.tails.flycount = set.BOOST_COUNT
	p.abilityused = false
end

-- Tails flight activated
function A_KRTailsFly(player)
	if not krework.checkExistence(player)
	or player.mo.skin != "tails"
	or player.bot
	or not (player.pflags & PF_JUMPED)
		return false
	end
		
	if player.krework.jumpused
		S_StartSound(mo, sfx_pudpud)
		return true
	end
	
	local mo = player.mo
	
	if (mo.krework.tails.flytime > 0 and mo.krework.tails.flytime < set.FLY_SPAMLOCK)
		return true
	end
	
	local speed = FixedHypot(mo.momx, mo.momy)
	local momangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	
	-- Momentum
	mo.momz = $ / 2
	P_Thrust(mo, momangle + ANGLE_180, speed / 4)
	
	-- Flags, states
	mo.krework.aircurled = false
	mo.krework.tails.flytime = 0
	mo.krework.tails.flycount = (mo.eflags & MFE_UNDERWATER) and set.BOOST_COUNT or ($1 - 1)
	player.pflags = $ | PF_NOJUMPDAMAGE
	player.pflags = $ & ~PF_SPINNING
	mo.state = S_PLAY_FLY
	
	-- Sound effect
	if not (mo.eflags & MFE_UNDERWATER)
		S_StartSound(mo, sfx_putput)
	end
	return true
end

function A_KRTailsHover(player)
	local mo = player.mo
	local time = mo.krework.tails.flytime

	-- Declare flight-related variables
	local hoverspeed = set.FLY_HOVERSPEED
	local flip = P_MobjFlip(mo)
	
	-- Change state to tired to show when you can boost again
	if time < set.FLY_SPAMLOCK
		player.krework.jumpused = true
		if mo.state == S_PLAY_FLY
			player.followmobj.tics = max($-1, 1)
		end
	end
	
	-- Change sound when you hit tired physics
	if time < set.FLY_TIRED_TIME
		if time % 10 == 1
		and (mo.state == S_PLAY_FLY or mo.state == S_PLAY_FLY_TIRED)
			S_StartSound(mo, sfx_s3kba)
		end
	else
		if mo.state == S_PLAY_FLY
			mo.state = S_PLAY_FLY_TIRED
		end
	
		if mo.state == S_PLAY_FLY_TIRED 
			if mo.eflags & MFE_UNDERWATER
				mo.state = S_PLAY_FLY
			elseif time % 5 == 1
				S_StartSound(mo, sfx_s3kbb)
			end
		end
		
		-- Start descending slightly faster if in the air
		if not (mo.eflags & MFE_UNDERWATER)
			hoverspeed = set.FLY_TIRED_HOVERSPEED
		end
	end
	
	-- Decide on vertboost duration
	local ascendtics = set.FLY_BOOST_TIME
	if mo.eflags & MFE_UNDERWATER
		ascendtics = set.SWIM_BOOST_TIME
	end
	local speed = FixedHypot(mo.momx, mo.momy)
	
	-- Fall no faster than your hover speed
	if (flip == 1 and mo.momz < hoverspeed) or (flip == -1 and mo.momz > -hoverspeed)
		P_SetObjectMomZ(mo, hoverspeed, false)
	end
	
	-- If you're still in vertboost, apply boost
	if (time <= ascendtics)
		-- Calculate boost to apply
		local ascendspeed = set.BOOST_UPSPEED
		if mo.eflags&MFE_UNDERWATER
			ascendspeed = $ * 2/3
		end
		
		P_SetObjectMomZ(mo, ascendspeed, true)
		
		-- Dust clouds (or bubbles) to represent Tails boosting
		if time % 1 == 0
			local source
			local dusttype
			local dust1x
			local dust1y
			local dust2x
			local dust2y
			local dustz
			local dustscale
			local dustcolor
			
			-- Spawn bubbles behind Tails's feet when swimming
			if mo.eflags&MFE_UNDERWATER
				source = player.mo
				
				-- Random deviation from center
				local bubblewobble = 2 * FRACUNIT / 5
				-- Need this because P_RandomRange can't handle a range >=65536
				local bubblewobblemult = 2
				
				local angle = (source.angle + ANGLE_157h)
				dust1x = FixedMul(cos(angle), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dust1y = FixedMul(sin(angle), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dust2x = FixedMul(cos(angle+ANGLE_45), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dust2y = FixedMul(sin(angle+ANGLE_45), source.radius) + P_RandomRange(0-bubblewobble, bubblewobble) * bubblewobblemult
				dustz = 0
				
				dusttype = MT_MEDIUMBUBBLE
				
				dustscale = false
				dustcolor = false
			-- Spawn dust clouds above Tails's tails when flying
			else
				source = player.followmobj
				if not source
					return
				end
				
				-- Tails's tails have a 4-frame animation that rotates 180 degrees
				local angle = (player.mo.angle + (source.frame * ANGLE_45))
				dust1x = FixedMul(cos(angle), source.radius)
				dust1y = FixedMul(sin(angle), source.radius)
				dust2x = FixedMul(cos(angle+ANGLE_180), source.radius)
				dust2y = FixedMul(sin(angle+ANGLE_180), source.radius)
				dustz = source.height*4/5
				
				dusttype = MT_SPINDUST
				
				dustscale = set.BOOST_SCALE_DUST
				dustcolor = set.BOOST_COLOR_DUST
			end
			
			local d1 = P_SpawnMobjFromMobj(source, dust1x, dust1y, dustz, dusttype)
			local d2 = P_SpawnMobjFromMobj(source, dust2x, dust2y, dustz, dusttype)
			
			-- Change color to represent spending boosts
			if dustcolor and mo.krework.tails.flycount < 3
				local dcolor = set.BOOST_DUST_COLORS[mo.krework.tails.flycount] or set.BOOST_DUST_DEFAULT
				d1.color, d2.color = dcolor, dcolor
				d1.colorized, d2.colorized = true, true
			end
			
			-- Reduce in size to represent spending boosts
			if dustscale
				d1.scale = $ * (mo.krework.tails.flycount+1) / 3
				d2.scale = d1.scale
			end
		end
	end
	
	-- Drop out of the air, keep swimming in water
	if time >= set.FLY_END_TIME and not (mo.eflags&MFE_UNDERWATER)
		mo.krework.tails.flytime = 0
		mo.state = S_PLAY_FALL
	else
		mo.krework.tails.flytime = $ + 1
	end
end
